#![feature(type_alias_impl_trait)]

use bcm2711_pac::gpio;
use bcm2711_pac::spi;
use tock_registers::interfaces::{ReadWriteable, Readable, Writeable};
use solid::{interrupt, singleton::pin_singleton, thread::CpuCx};
use std::sync::atomic::{AtomicBool, Ordering, AtomicU8, AtomicI32};

struct XferState {
	send_data: [AtomicU8; 2],
	recv_data: AtomicI32,
	is_first: AtomicBool,
	is_read: AtomicBool,
	is_done: AtomicBool,
}

static XFER_STATE: XferState = XferState{
	send_data: [AtomicU8::new(0),AtomicU8::new(0)],
	recv_data: AtomicI32::new(0),
	is_first:  AtomicBool::new(false),
	is_read:  AtomicBool::new(false),
	is_done:  AtomicBool::new(false)
};

/// アクティブなSPI転送の発行元タスクIDを格納するグローバル変数。
/// 制御機構を独占するためのミューテックスの役割も兼ねる。
static XFER_ACTIVE_TID: AtomicI32 = AtomicI32::new(0);

fn gpio_regs() -> &'static gpio::Registers {
   	// Safety: SOLID for RaPi4B provides an identity mapping in this area, and we don't alter
   	// the mapping
   	unsafe { &*(gpio::BASE.to_arm_pa().unwrap() as usize as *const gpio::Registers) }
}

fn spi_regs() -> &'static spi::Registers {
   	// Safety: SOLID for RaPi4B provides an identity mapping in this area, and we don't alter
   	// the mapping
   	unsafe { &*(spi::BASE_SPI0.to_arm_pa().unwrap() as usize as *const spi::Registers) }
}

fn gpio_init() {
   	for pin in 8..=11 {
       	gpio_regs().gpfsel[pin / gpio::GPFSEL::PINS_PER_REGISTER].modify(
           	gpio::GPFSEL::pin(pin % gpio::GPFSEL::PINS_PER_REGISTER).val(gpio::GPFSEL::ALT0),
       	);
   	}
}

fn spi_trans_handler() {
	//DONEビットが1でFIFO書き込みデータがあるなら、送信FIFOに書き込みデータをライトする。
	//DONEビットが1でFIFO書き込みデータがないなら、受信FIFOからデータをリードし割り込み発生フラグを1にするClearTA
	//DONEビットが0であれば何もしない
	if spi_regs().cs.is_set(spi::CS::DONE) {
		let transfer_type = XFER_STATE.is_read.load(Ordering::SeqCst);
		let first_flag = XFER_STATE.is_first.load(Ordering::SeqCst);

		if transfer_type == true && first_flag == true //regread初回割り込み
		{
			spi_regs().fifo.write(spi::FIFO::DATA.val(XFER_STATE.send_data[0].load(Ordering::SeqCst).into()));
			spi_regs().fifo.write(spi::FIFO::DATA.val(0));
			spi_regs().fifo.write(spi::FIFO::DATA.val(0));
			XFER_STATE.is_first.store(false, Ordering::SeqCst);
		}
		else if transfer_type == false && first_flag == true //regwrite初回割り込み
		{
			spi_regs().fifo.write(spi::FIFO::DATA.val(XFER_STATE.send_data[0].load(Ordering::SeqCst).into()));
			spi_regs().fifo.write(spi::FIFO::DATA.val(XFER_STATE.send_data[1].load(Ordering::SeqCst).into()));
			XFER_STATE.is_first.store(false, Ordering::SeqCst);
		}
		else 
		{
			if transfer_type == true //regread: 3回転送分のデータがRX FIFOに入っているという決め打ち。(Dummy, 下位1バイト、上位1バイト）
			{
				let retdata_h: u32;
				let retdata_l: u32;

				spi_regs().fifo.read(spi::FIFO::DATA); //Dummy read

				retdata_l = spi_regs().fifo.read(spi::FIFO::DATA);
				retdata_h = spi_regs().fifo.read(spi::FIFO::DATA);
				XFER_STATE.recv_data.store(i32::from(i16::from_le_bytes([retdata_l as u8, retdata_h as u8])),Ordering::SeqCst);
			}

			spi_regs().cs.modify(spi::CS::TA::CLEAR);
			XFER_STATE.is_done.store(true, Ordering::SeqCst);
			let tid = XFER_ACTIVE_TID.load(Ordering::Acquire);
			if tid != 0 {
				unsafe { itron::abi::wup_tsk(tid) };
			}
		}
	}
	else {}
}

fn spi_init() {
   	const CLKSET: u32 = 0x00cb;

   	gpio_init();

   	//SPIのCLKレジスタにSPI周波数を設定
   	spi_regs().clk.write(spi::CLK::CDIV.val(CLKSET));

   	//CS : 00 = Chip select 0  -> CS.bit0 and 1
   	//CPOL: Clock Polarity 1 = Rest state of clock = High -> CS.bit3
   	//CPHA: Clock Phase 1 = First SCLK transition at beginning of data bit. -> CS.bit2
   	spi_regs().cs.modify(spi::CS::CS::ChipSelect0);
   	spi_regs().cs.modify(spi::CS::CPOL::RestStateIsHigh);
   	spi_regs()
       	.cs
       	.modify(spi::CS::CPHA::FirstSclkTransitionAtBeginningOFDataBit);

	println!("Register SPI Interrupt.");
	let max_priority: i32 = unsafe{solid::abi::SOLID_INTC_GetPriorityLevel().into()};
	let spi_handler_option = interrupt::HandlerOptions::new(interrupt::Number(150),  max_priority - 2);
	let spi_handler = pin_singleton!(: Handler<_> = interrupt::Handler::new(|_: CpuCx<'_>| spi_trans_handler())).unwrap(); 
    
	assert!(
		spi_handler.register_static(&spi_handler_option)
		.expect("unable to register interrupt handler"),
		"interrupt handler was already registered"
	);

	spi_regs().cs.modify(spi::CS::INTD::SET);  //割り込み発生要因：DONE = 1

	interrupt::Number(150).enable().expect("unable to enable interrupt line");

}

pub fn regwrite(regaddr: u8, writeval: u8) {
	//(1)呼出し元タスクIDを取得しグローバル変数に格納する
	let mut tid = 0;
	assert_eq!(unsafe { itron::abi::get_tid(&mut tid) }, itron::abi::E_OK);
	assert_ne!(tid, 0);

	XFER_ACTIVE_TID.compare_exchange(0, tid, Ordering::AcqRel, Ordering::Relaxed).expect("transfer already in progress");

	//(2)転送完了フラグをクリアし、FIFO書き込みデータ準備
	XFER_STATE.is_done.store(false, Ordering::SeqCst);
	XFER_STATE.send_data[0].store(regaddr, Ordering::SeqCst);
	XFER_STATE.send_data[1].store(writeval, Ordering::SeqCst);
	XFER_STATE.is_first.store(true, Ordering::SeqCst);
	XFER_STATE.is_read.store(false, Ordering::SeqCst);

	//(3)CSレジスタのCLEARビットに0x03を書いてTX-FIFO, RX-FIFOクリア
   	spi_regs().cs.modify(spi::CS::CLEAR_TX::SET + spi::CS::CLEAR_RX::SET);

	//(4)CSレジスタのTAビットを1にする。
   	spi_regs().cs.modify(spi::CS::TA::SET);

   	//(5)転送完了割り込み発生フラグがtrueになるまで待つ
   	while !XFER_STATE.is_done.load(Ordering::Acquire){
		assert_eq!(unsafe { itron::abi::slp_tsk() }, itron::abi::E_OK);
	}
	//(6)タスクIDを取り除く
	XFER_ACTIVE_TID.store(0, Ordering::Release);
}

fn regread(regval: u8) -> i32 {
	//(1)呼出し元タスクIDを取得しグローバル変数に格納する
	let mut tid = 0;
	assert_eq!(unsafe { itron::abi::get_tid(&mut tid) }, itron::abi::E_OK);
	assert_ne!(tid, 0);
		
	XFER_ACTIVE_TID.compare_exchange(0, tid, Ordering::AcqRel, Ordering::Relaxed).expect("transfer already in progress");

	//(2)転送完了フラグをクリアし、FIFO書き込みデータ準備
	XFER_STATE.is_done.store(false, Ordering::SeqCst);
	XFER_STATE.send_data[0].store(regval, Ordering::SeqCst);
	XFER_STATE.is_first.store(true, Ordering::SeqCst);
	XFER_STATE.is_read.store(true, Ordering::SeqCst);

   	//(3)CSレジスタのCLEARビットに0x03を書いてTX-FIFO, RX-FIFOクリア
   	spi_regs().cs.modify(spi::CS::CLEAR_TX::SET + spi::CS::CLEAR_RX::SET);

	//(4)CSレジスタのTAビットを1にする。
   	spi_regs().cs.modify(spi::CS::TA::SET);

   	//(5)転送完了割り込み発生フラグがtrueになるまで待つ
   	while !XFER_STATE.is_done.load(Ordering::Acquire){
		assert_eq!(unsafe { itron::abi::slp_tsk() }, itron::abi::E_OK);
	}

    //(6)タスクIDを取り除く
	XFER_ACTIVE_TID.store(0, Ordering::Release);

	XFER_STATE.recv_data.load(Ordering::SeqCst)

}

pub fn init() {
	const BW_RATE: u8 = 0x2c;
   	const BW_RATE_VAL: u8 = 0x0b;
   	const DATA_FORMAT: u8 = 0x31;
   	const DATA_FORMAT_VAL: u8 = 0x0B; //(4 mg/LSB +-16g)
   	const POWER_CTL: u8 = 0x2D;
   	const POWER_CTL_VAL: u8 = 0x08;

	spi_init();

	//ADXL345_init
   	regwrite(BW_RATE, BW_RATE_VAL);
   	regwrite(DATA_FORMAT, DATA_FORMAT_VAL);
   	regwrite(POWER_CTL, POWER_CTL_VAL);
}

//配列、、、で書いてみる。
//
pub fn current_acceleration_array() -> [f32; 3] {
	const DATAX0: u8 = 0x32;
	const DATAY0: u8 = 0x34;
	const DATAZ0: u8 = 0x36;

	let mut axis_array: [f32; 3] = [0.0, 0.0, 0.0];

   	//連続複数リードを行う。-> 0xc0とORをとる。
 	//0.0392266=(4/1000*9.80665)
	axis_array[0] = regread(0xc0 | DATAX0) as f32 * 0.0392266;
	axis_array[1] = regread(0xc0 | DATAY0) as f32 * 0.0392266;
	axis_array[2] = regread(0xc0 | DATAZ0) as f32 * 0.0392266;

	return axis_array;
}

//タプル、、、で書いてみる。
//
pub fn current_acceleration_tuple() -> (f32, f32, f32) {
	const DATAX0: u8 = 0x32;
	const DATAY0: u8 = 0x34;
	const DATAZ0: u8 = 0x36;

   	//連続複数リードを行う。-> 0xc0とORをとる。
 	//0.0392266=(4/1000*9.80665)
	let x_axis = regread(0xc0 | DATAX0) as f32 * 0.0392266;
	let y_axis = regread(0xc0 | DATAY0) as f32 * 0.0392266;
	let z_axis = regread(0xc0 | DATAZ0) as f32 * 0.0392266;

	return (x_axis, y_axis, z_axis);
}
